import os
import argparse
from cnorm.parsing.declaration import Declaration
from cnorm.passes import to_c
from pyrser.passes import to_yml

parser = argparse.ArgumentParser(prog="cnorm", description="CNORM a C language front-end in python")
parser.add_argument("filenames", help="process FILE with cnorm",
                    metavar="FILE", type=str, nargs='+')
parser.add_argument("-y", "--yml", dest="yml", help="show AST nodes as yml", action='store_true')
parser.add_argument("-d", "--dump", dest="dump", help="show AST nodes as vars", action='store_true')

args = parser.parse_args()

for f in args.filenames:
    if f != '-' and os.path.exists(f):
        cparse = Declaration()
        ast = cparse.parse_file(f)
        if args.dump:
            print(vars(ast))
        if args.yml:
            print(ast.to_yml())
        if ast:
            print(ast.to_c())
        else:
            print("something goes wrong")



from pyrser import meta, directives
from pyrser.grammar import Grammar
from cnorm import nodes

class   Literal(Grammar):
    """
        interaction with other CNORM PART:

            Expression.primary_expression -> Literal.literal
    """

    grammar = """
        dot ::= '.' !'.'
        ;

        pow ::= 'p' | 'P'
        ;

        exp ::= 
            ['e'|'E'] ['+'|'-']? ['0'..'9']+
        ;

        unsigned_suffix ::=
            ['u'|'U']
        ;

        long_suffix ::=
            ['l'|'L']
        ;

        float_suffix ::=
            ['f'|'F']
        ;

        complex_suffix ::=
            ['i'|'I'|'j'|'J']
        ;

        decimal_const ::=
            ['0'..'9']+
            unsigned_suffix?
            long_suffix?
            long_suffix?
        ;

        hexadecimal_prefix ::= '0' ['x'|'X']
        ;

        hexadecimal_digit ::= ['0'..'9'|'a'..'f'|'A'..'F']
        ;

        hexadecimal_const_int ::=
            hexadecimal_prefix [hexadecimal_digit]+
            unsigned_suffix?
            long_suffix?
            long_suffix?
        ;

        octal_digit ::= '0'..'7'
        ;

        octal_const ::= 
            '0' octal_digit+
            [
               dot octal_digit+
               [pow ['+'|'-']? decimal_const]?
            ]?
        ;

        double_const ::=
            [decimal_const dot ['0'..'9']*| dot ['0'..'9']+] exp?
            long_suffix?
            float_suffix?
            complex_suffix?
        ;

        encoding_prefix ::= "u8" | 'u' | 'U' | 'L'
        ;

        string_const ::=
            encoding_prefix? Base.string @ignore("C/C++") Base.string*
        ;

        char_const ::=
            encoding_prefix? Base.char
        ;

        literal ::= @ignore('null')
            [
                hexadecimal_const_int
                | octal_const
                | double_const
                | decimal_const
                | string_const
                | char_const
            ]:val
            #new_literal(_, val)
        ;
    """

    def parse(self, source, entry):
        res = Grammar.parse(self, source, entry)
        if hasattr(res, 'node'):
            return res.node
        return res

@meta.hook(Literal)
def new_literal(self, ast, val):
    ast.node = nodes.Literal(val.value)
    return True



from cnorm.nodes import Decl, Stmt, Expr
from cnorm.nodes import DeclType, PointerType, ArrayType, ParenType, FuncType, PrimaryType

__all__ = [
    'Decl',
    'Stmt',
    'Expr',
    'DeclType',
    'PointerType',
    'ArrayType',
    'ParenType',
    'FuncType',
    'PrimaryType',
    'passes',
    'type_handling',
    ]



from pyrser import meta, directives
from pyrser.grammar import Grammar
from pyrser.hooks import copy, echo
from pyrser.parsing.node import Node
from pyrser.directives import ignore
from cnorm import nodes
from cnorm.parsing.expression import Expression

class Statement(Grammar, Expression):
    """
        interaction with other CNORM PART:

        Declaration.init_declarator -> compound_statement
        Expression.primary_expression -> block_item_list
    """
    entry = "single_statement"
    grammar = """

        /*
            Comment works as in C/C++
        */

        single_statement ::=
            compound_statement:_
            | labeled_statement:_
            | expression_statement:_
        ;

        compound_statement ::=
            [
            '{' 
                "":current_block
                #new_blockstmt(_, current_block)
                [
                    line_of_code
                ]*
            '}'
            ]
        ;

        line_of_code ::=
            single_statement:line
            #end_loc(current_block, line)
        ;

        labeled_statement ::=
            rootidentifier:id
            [
            #check_stmt(id, "if") if_statement:_
            | #check_stmt(id, "for") for_statement:_
            | #check_stmt(id, "while") while_statement:_
            | #check_stmt(id, "switch") switch_statement:_
            | #check_stmt(id, "do") do_statement:_
            | #check_stmt(id, "return") return_statement:_
            | #check_stmt(id, "goto") goto_statement:_
            | #check_stmt(id, "case") case_statement:_
            | #check_stmt(id, "break") ';' #new_break(_)
            | #check_stmt(id, "continue") ';' #new_continue(_)
            | ':' #new_label(_, id)
            ]
        ;

        if_statement ::=
            '(' expression:cond ')'
            single_statement:then
            ["else"
            // TODO: explain variable forwarding of 'else'
            single_statement:else 
            ]?:else
            #new_if(_, cond, then, else)
        ;

        for_statement ::=
            '('
                expression_statement:init
                expression_statement:cond
                expression?:inc
            ')'
            single_statement:body
            #new_for(_, init, cond, inc, body)
        ;

        while_statement ::=
            '(' expression:cond ')'
            single_statement:body
            #new_while(_, cond, body)
        ;

        switch_statement ::=
            '(' expression:cond ')'
            single_statement:body
            #new_switch(_, cond, body)
        ;

        do_statement ::=
            single_statement:body
            "while" '(' expression:cond ')' ';'
            #new_do(_, cond, body)
        ;

        return_statement ::=
            expression?:e ';'
            #new_return(_, e)
        ;

        goto_statement ::=
            expression:e ';'
            #new_goto(_, e)
        ;

        range_expression ::=
            constant_expression:_ ["..." constant_expression:r #new_range(_, r) ]?
        ;

        case_statement ::=
            range_expression:e #new_case(_, e)
            ':'
        ;

        expression_statement ::=
            [expression:e #new_expr(_, e)]?
            ';'
        ;

    """

    def parse(self, source, entry=None):
        res = Grammar.parse(self, source, entry)
        if hasattr(res, 'node'):
            return res.node
        return res

@meta.hook(Statement)
def new_expr(self, ast, expr):
    ast.node = nodes.ExprStmt(expr.node)
    return True

@meta.hook(Statement)
def new_if(self, ast, cond_expr, then_expr, else_expr):
    thenbody = None
    if hasattr(then_expr, 'node'):
        thenbody = then_expr.node
    if hasattr(else_expr, 'node'):
        ast.node = nodes.If(cond_expr.node, thenbody, else_expr.node)
    else:
        ast.node = nodes.If(cond_expr.node, thenbody)
    return True

@meta.hook(Statement)
def new_for(self, ast, init, cond, inc, body):
    init_body = None
    if hasattr(init, 'node'):
        init_body = init.node
    cond_body = None
    if hasattr(cond, 'node'):
        cond_body = cond.node
    inc_body = None
    if hasattr(inc, 'node'):
        inc_body = inc.node
    body_body = None
    if hasattr(body, 'node'):
        body_body = body.node
    ast.node = nodes.For(init_body, cond_body, inc_body, body_body)
    return True

@meta.hook(Statement)
def new_while(self, ast, cond, body):
    thebody = None
    if hasattr(body, 'node'):
        thebody = body.node
    ast.node = nodes.While(cond.node, thebody)
    return True

@meta.hook(Statement)
def new_switch(self, ast, cond, body):
    thebody = None
    if hasattr(body, 'node'):
        thebody = body.node
    ast.node = nodes.Switch(cond.node, thebody)
    return True

@meta.hook(Statement)
def new_do(self, ast, cond, body):
    thebody = None
    if hasattr(body, 'node'):
        thebody = body.node
    ast.node = nodes.Do(cond.node, thebody)
    return True

@meta.hook(Statement)
def new_return(self, ast, expr):
    theexpr = None
    if hasattr(expr, 'node'):
        theexpr = expr.node
    ast.node = nodes.Return(theexpr)
    return True

@meta.hook(Statement)
def new_goto(self, ast, expr):
    ast.node = nodes.Goto(expr.node)
    return True

@meta.hook(Statement)
def new_range(self, ast, expr):
    ast.node = nodes.Range(nodes.Raw('...'), [ast.node, expr.node])
    return True

@meta.hook(Statement)
def new_case(self, ast, expr):
    ast.node = nodes.Case(expr.node)
    return True

@meta.hook(Statement)
def new_break(self, ast):
    ast.node = nodes.Break()
    return True

@meta.hook(Statement)
def new_continue(self, ast):
    ast.node = nodes.Continue()
    return True

@meta.hook(Statement)
def new_label(self, ast, ident):
    ast.node = nodes.Label(ident.value)
    return True

@meta.hook(Statement)
def new_blockstmt(self, ast, current_block):
    ast.node = nodes.BlockStmt([])
    current_block.node = ast.node
    parent = self.rulenodes.parents
    if 'current_block' in parent:
        current_block.node.types = parent['current_block'].node.types.new_child()
    return True

@meta.hook(Statement)
def end_loc(self, current_block, ast):
    line = None
    if hasattr(ast, 'node'):
        line = ast.node
    current_block.node.body.append(line)
    return True

@meta.hook(Statement)
def check_stmt(self, ident, val):
    return ident.value == val



#!/usr/bin/env python3

"""
        Test Basic parsing for a little language
"""
import sys
from pyrser import grammar, meta

class   TypeLang(grammar.Grammar):
        grammar = """
        main ::= [id:id ]+
        ;
        """
        entry = "main"

if __name__ == "__main__":
        print("DEBUT:%s" % sys.path)
        oAst = {}
        print("DEBUT2")
        try:
                oLang = TypeLang()
        except IOError as e:
                print("Detail:%s" % repr(e))
        print("DEBUT3")
        oAst = oLang.parse("""
                cool ca marche
                """)
        print("DEBUT4")
        print(repr(oAst))



from pyrser import meta, directives
from pyrser.passes import to_yml
from pyrser.hooks import copy, echo
from pyrser.parsing.node import Node
from cnorm import nodes
from pyrser.grammar import Grammar
from cnorm.parsing.statement import Statement
from cnorm.parsing.expression import Idset
from weakref import ref

import sys

sys.setrecursionlimit(10000)

class Declaration(Grammar, Statement):
    """
        interaction with other CNORM PART:

    """
    #: entry point for C programming language
    entry = "translation_unit"

    #: complete C declaration grammar
    grammar = """

        translation_unit ::=
            @ignore("C/C++")
            [
                "":current_block
                #new_root(_, current_block)
                [
                    declaration
                ]*
            ]
            Base.eof
        ;

        declaration ::=
            ';' // garbage single comma
            |
            c_decl
            |
            preproc_decl
            |
            asm_decl
        ;

        preproc_decl ::=
            ['#' preproc_directive ]:decl
            #raw_decl(decl)
            #end_decl(current_block, decl)
        ;

        asm_decl ::=
            [
            Base.id:i
            #check_asm(i)
            [
                Base.id:i
                #check_quali(i)
            ]?
            attr_asm_decl_follow
            ';'?
            ]:decl
            #raw_decl(decl)
            #end_decl(current_block, decl)
        ;

        attr_asm_decl ::=
            [
            Base.id:i
            #check_asmattr(i)
            attr_asm_decl_follow
            ]:_
        ;

        attr_asm_decl_follow ::=
            '(' dummy_with_paren* ')' 
            | '{' dummy_with_brace* '}'
            | '__extension__'
        ;

        c_decl ::=
            "":local_specifier
            #create_ctype(local_specifier)
            declaration_specifier*:dsp
            init_declarator:decl
            #not_empty(current_block, dsp, decl)
            #end_decl(current_block, decl)
            [
                ',' 
                #copy_ctype(local_specifier, decl)
                init_declarator:decl
                #end_decl(current_block, decl)
            ]*
            [
                ';'
                |
                Statement.compound_statement:b
                #add_body(decl, b)
            ]
        ;

        declaration_specifier ::=
            Base.id:i
            #new_decl_spec(local_specifier, i, current_block)
            [
                #is_composed(local_specifier)
                composed_type_specifier
                |
                #is_enum(local_specifier)
                enum_specifier
                |
                #is_typeof(i)
                typeof_expr
            ]?
            |
            attr_asm_decl:attr
            #add_attr_specifier(local_specifier, attr)
        ;

        type_qualifier ::=
            Base.id:i
            #add_qual(local_specifier, i)
            |
            attr_asm_decl:attr
            #add_attr_specifier(local_specifier, attr)
        ;

        name_of_composed_type ::= Base.id ;
        composed_type_specifier ::=
            [
                attr_asm_decl:attr
                #add_attr_composed(local_specifier, attr)
            ]?
            name_of_composed_type?:n
            composed_fields?:body
            #add_composed(local_specifier, n, body)
        ;

        composed_fields ::=
            '{'
                "":current_block
                #new_composed(_, current_block)
                declaration*
            '}'
        ;

        enum_name ::= Base.id ;
        enum_specifier ::=
            enum_name?:n
            enumerator_list?:body
            #add_enum(local_specifier, n, body)
        ;

        enumerator_list ::=
            '{'
                enumerator:e
                #add_enumerator(_, e)
                [
                    ',' enumerator:e
                    #add_enumerator(_, e)
                ]*
                ','? // trailing comma
            '}'
        ;

        enumerator ::=
            identifier:i ['=' constant_expression:c]?:c
            #new_enumerator(_, i, c)
        ;

        typeof_expr ::=
            // TODO: split inside typeof
            ['('
                [
                    type_name !!')'
                    | expression
                ]
            ')']:tof
            #add_typeof(local_specifier, tof)
        ;

        init_declarator ::=
            declarator:_
            [
                ':'
                constant_expression:cexpr
                #colon_expr(_, cexpr)
            ]?
            [
                attr_asm_decl:attr
                #add_attr_decl(_, attr)
            ]*
            [
                '='
                initializer:aexpr
                #assign_expr(_, aexpr)
            ]?
            !![','|';'|'{']
        ;

        declarator ::=
            [
                "*"
                #first_pointer(local_specifier) 
                declarator_recurs:_
                | 
                absolute_declarator:_
            ]
            #commit_declarator(_, local_specifier)
        ;

        declarator_recurs ::=
            pointer absolute_declarator:_
        ;

        pointer ::=
            [
                "*" #add_pointer(local_specifier)
                | type_qualifier
            ]*
        ;

        f_or_v_id ::= identifier;
        absolute_declarator ::=
                [
                    '('
                        #add_paren(local_specifier)
                        type_qualifier?
                        declarator_recurs:_
                        #close_paren(local_specifier)
                    ')'
                    | 
                    f_or_v_id?:name
                    #name_absdecl(local_specifier, name)
                ]
                direct_absolute_declarator?
        ;

        direct_absolute_declarator ::=
            [
                '['
                    // TODO: handle c99 qual for trees
                    "static"?
                    ["const"|"volatile"]?
                    "static"?
                    [
                        assignement_expression:e
                        | '*':star #new_raw(e, star)
                    ]?:e
                    #add_ary(local_specifier, e)
                ']'
            ]+
            |
                '('
                #open_params(local_specifier)
                [
                    //kr_parameter_type_list
                    //| 
                    parameter_type_list
                ]?
                ')'
            /*
            [ // K&R STYLE
                !![';'|','|'{'|'('|')']
                | declaration*
            ]
            */
        ;

        kr_parameter_type_list ::=
            identifier [',' identifier]* !!')'
        ;

        parameter_type_list ::=
            [type_name ';']*
            [
                parameter_list
            ]?
            ','?
            ["..." #add_ellipsis(local_specifier)]?
        ;

        parameter_list ::=
            parameter_declaration:p
            #add_param(local_specifier, p)
            [','
                parameter_declaration:p
                #add_param(local_specifier, p)
            ]*
        ;

        parameter_declaration ::=
            type_name:_
        ;

        initializer ::=
            initializer_block:_
            | assignement_expression:_
        ;
        
        initializer_block ::=
            '{'
                "":init_list
                #new_blockinit(init_list)
                [initializer_list]?
                ','? // trailing comma
                #copy(_, init_list)
            '}'
        ;

        initializer_list ::=
            designation?:dsign
            initializer:init
            #add_init(init_list, init, dsign)
            [
                ','
                designation?:dsign
                initializer:init
                #add_init(init_list, init, dsign)
            ]*
        ;

        designation ::=
            designation_list+ '='?
            | identifier ':'
        ;

        designation_list ::=
            '['
                range_expression
            ']'
            | dot identifier
        ;

        type_name ::=
            "":local_specifier
            #create_ctype(local_specifier)
            declaration_specifier+ declarator:_
        ;

        ///////// OVERLOAD OF STATEMENT
        // add declaration in block
        line_of_code ::=
                    declaration
                |
                    single_statement:line
                    #end_loc(current_block, line)
        ;

        for_statement ::=
            '('
                [
                    "":current_block
                    #for_decl_begin(current_block)
                    declaration
                    #for_decl_end(init, current_block)
                |expression_statement:init]:init
                expression_statement:cond
                expression?:inc
            ')'
            single_statement:body
            #new_for(_, init, cond, inc, body)
        ;

        ///////// OVERLOAD OF EXPRESSION
        // add cast / sizeof
        unary_expression ::=
            // CAST
            '(' type_name:t ')'
            [
                // simple cast
                unary_expression:_
                |
                // compound literal
                initializer_block:_
            ]
            #to_cast(_, t)
            | // SIZEOF
            Base.id:i #sizeof(i)
            [
                '(' type_name:n ')'
                | Expression.unary_expression:n
            ]:n
            #new_sizeof(_, i, n)
            | Expression.unary_expression:_
        ;

        // ({}) and __builtin_offsetof
        primary_expression ::=
            "({"
                "":current_block
                #new_blockexpr(_, current_block)
                [
                    line_of_code
                ]*
            "})"
            | // TODO: create special node for that
                "__builtin_offsetof"
                '(' [type_name ',' postfix_expression]:bof ')'
                #new_builtoffset(_, bof)
            |
            Expression.primary_expression:_
        ;

    """

    def after_parse(self, res):
        return res.node

@meta.hook(Declaration)
def check_asm(self, ident):
    if ident.value in Idset and Idset[ident.value] == "asm":
        return True
    return False

@meta.hook(Declaration)
def check_quali(self, ident):
    if ident.value in Idset and Idset[ident.value] == "qualifier":
        return True
    return False

@meta.hook(Declaration)
def check_asmattr(self, ident):
    if ident.value in Idset and (Idset[ident.value] == "asm" or \
        Idset[ident.value] == "attribute"):
        return True
    return False

@meta.hook(Declaration)
def new_root(self, ast, current_block):
    ast.node = nodes.RootBlockStmt([])
    current_block.node = ast.node
    return True

@meta.rule(Declaration, "preproc_directive")
def preproc_directive(self) -> bool:
    """Consume a preproc directive."""
    self._stream.save_context()
    if self.read_until("\n", '\\'):
        return self._stream.validate_context()
    return self._stream.restore_context()

@meta.hook(Declaration)
def raw_decl(self, decl):
    decl.node = nodes.Raw(decl.value)
    return True

@meta.hook(Declaration)
def create_ctype(self, lspec):
    lspec.ctype = None
    return True

@meta.hook(Declaration)
def copy_ctype(self, lspec, previous):
    lspec.ctype = previous.node.ctype.copy()
    return True


@meta.hook(Declaration)
def new_decl_spec(self, lspec, i, current_block):
    idsetval = ""
    if i.value in Idset:
        idsetval = Idset[i.value]
    # don't fuck with reserved keywords
    if idsetval == "reserved":
        return False
    # not for asm or attribute
    if idsetval != "" and idsetval[0] != 'a':
        lspec.ctype = nodes.makeCType(i.value, lspec.ctype)
        return True
    if hasattr(current_block.node, 'types') \
        and i.value in current_block.node.types:
        if lspec.ctype == None:
            lspec.ctype = nodes.PrimaryType(i.value)
        else:
            lspec.ctype._identifier = i.value
        lspec.ctype._identifier = i.value
        #print("type found <%s> %d" % (i.value, lspec.ctype._storage))
        return True
    return False

@meta.hook(Declaration)
def add_body(self, ast, body):
    ast.node.body = body.node
    return True

@meta.hook(Declaration)
def end_decl(self, current_block, ast):
    current_block.node.body.append(ast.node)
    if hasattr(ast.node, 'ctype') and ast.node._name != "" and \
        ast.node.ctype._storage == nodes.Storages.TYPEDEF:
        #print("new type %s" % ast.node._name)
        current_block.node.types[ast.node._name] = ref(ast.node)
    return True

@meta.hook(Declaration)
def not_empty(self, current_block, dsp, decl):
    # empty declspec only in global scope
    if type(current_block.node) is nodes.BlockStmt and dsp.value == "":
        return False
    return True

@meta.hook(Declaration)
def colon_expr(self, ast, expr):
    ast.node.colon_expr(expr.node)
    return True

@meta.hook(Declaration)
def assign_expr(self, ast, expr):
    ast.node.assign_expr(expr.node)
    return True

@meta.hook(Declaration)
def is_composed(self, lspec):
    if lspec.ctype._specifier == nodes.Specifiers.STRUCT or\
        lspec.ctype._specifier == nodes.Specifiers.UNION:
            return True
    return False

@meta.hook(Declaration)
def is_enum(self, lspec):
    if lspec.ctype._specifier == nodes.Specifiers.ENUM:
        return True
    return False

@meta.hook(Declaration)
def is_typeof(self, i):
    if i.value in Idset and Idset[i.value] == "typeof":
        return True
    return False

@meta.hook(Declaration)
def add_typeof(self, lspec, tof):
    lspec.ctype = nodes.PrimaryType("typeof" + tof.value)
    return True

@meta.hook(Declaration)
def add_qual(self, lspec, qualspec):
    dspec = qualspec.value
    if dspec in Idset and Idset[dspec] == "qualifier":
        cleantxt = dspec.strip("_")
        lspec.ctype.push(nodes.QualType(nodes.Qualifiers.map[cleantxt.upper()]))
        return True
    return False

@meta.hook(Declaration)
def add_attr_specifier(self, lspec, attrspec):
    if lspec.ctype == None:
        lspec.ctype = nodes.makeCType('int', lspec.ctype)
    lspec.ctype.push(nodes.AttrType(attrspec.value))
    return True

@meta.hook(Declaration)
def add_attr_composed(self, lspec, attrspec):
    if not hasattr(lspec.ctype, '_attr_composed'):
        lspec.ctype._attr_composed = []
    lspec.ctype._attr_composed.append(attrspec.value)
    return True

@meta.hook(Declaration)
def add_attr_decl(self, lspec, attrspec):
    if not hasattr(lspec.node, '_attr_decl'):
        lspec.node._attr_decl = []
    lspec.node._attr_decl.append(attrspec.value)
    return True

@meta.hook(Declaration)
def add_composed(self, lspec, n, block):
    ctype = nodes.ComposedType(n.value)
    if lspec.ctype != None:
        ctype._storage = lspec.ctype._storage
        #print("STORAGE %d" % ctype._storage)
        ctype._specifier = lspec.ctype._specifier
        if hasattr(lspec.ctype, '_attr_composed'):
            ctype._attr_composed = lspec.ctype._attr_composed
    lspec.ctype = ctype
    if hasattr(block, 'node'):
        lspec.ctype.fields = block.node.body
    return True

@meta.hook(Declaration)
def add_enum(self, lspec, n, block):
    ctype = nodes.ComposedType(n.value)
    if lspec.ctype != None:
        ctype._storage = lspec.ctype._storage
        ctype._specifier = lspec.ctype._specifier
    lspec.ctype = ctype
    if hasattr(block, 'node'):
        lspec.ctype.enums = block.node
    return True

@meta.hook(Declaration)
def add_enumerator(self, ast, enum):
    if not hasattr(ast, 'node'):
        ast.node = []
    ast.node.append(enum.node)
    return True

@meta.hook(Declaration)
def new_enumerator(self, ast, ident, constexpr):
    expr = None
    if hasattr(constexpr, 'node'):
        expr = constexpr.node
    ast.node = nodes.Enumerator(ident.value, expr)
    return True

@meta.hook(Declaration)
def new_composed(self, ast, current_block):
    ast.node = nodes.BlockStmt([])
    current_block.node = ast.node
    parent = self.rulenodes.parents
    if 'current_block' in parent:
        current_block.node.types = parent['current_block'].node.types.new_child()
    return True

@meta.hook(Declaration)
def first_pointer(self, lspec):
    if not hasattr(lspec, 'ctype') or lspec.ctype == None:
        lspec.ctype = nodes.makeCType('int', lspec.ctype)
    lspec.ctype.push(nodes.PointerType())
    return True

@meta.hook(Declaration)
def commit_declarator(self, ast, lspec):
    if hasattr(lspec.ctype, '_params'):
        lspec.ctype.__class__ = nodes.FuncType
    name = ""
    if hasattr(lspec, '_name'):
        name = lspec._name
    ast.node = nodes.Decl(name, lspec.ctype)
    return True

@meta.hook(Declaration)
def add_pointer(self, lspec):
    if not hasattr(lspec, 'ctype'):
        lspec.ctype = nodes.makeCType('int', lspec.ctype)
    if not hasattr(lspec.ctype, 'push'):
        return False
    lspec.ctype.push(nodes.PointerType())
    return True

@meta.hook(Declaration)
def add_paren(self, lspec):
    if not hasattr(lspec, 'ctype') or lspec.ctype == None:
        lspec.ctype = nodes.makeCType('int')
    paren = nodes.ParenType()
    if not hasattr(lspec, 'cur_paren'):
        lspec.cur_paren = []
        lspec.cur_paren.append(ref(lspec.ctype))
    last = lspec.cur_paren.pop()
    lspec.cur_paren.append(ref(paren))
    lspec.cur_paren.append(last)
    lspec.ctype.push(paren)
    return True

@meta.hook(Declaration)
def add_ary(self, lspec, expr):
    if not hasattr(lspec, 'ctype') or lspec.ctype == None:
        lspec.ctype = nodes.makeCType('int')
    aryexpr = None
    if hasattr(expr, 'node'):
        aryexpr = expr.node
    if not hasattr(lspec, 'cur_paren'):
        lspec.cur_paren = []
        lspec.cur_paren.append(ref(lspec.ctype))
    lspec.cur_paren[-1]().push(nodes.ArrayType(aryexpr))
    return True

@meta.hook(Declaration)
def name_absdecl(self, ast, ident):
    if ident.value != "":
        ast._name = ident.value
        ast._could_be_fpointer = True
    return True

@meta.hook(Declaration)
def close_paren(self, lspec):
    lspec.cur_paren.pop()
    return True

@meta.hook(Declaration)
def open_params(self, lspec):
    if lspec.ctype == None:
        lspec.ctype = nodes.makeCType('int')
    if not hasattr(lspec, 'cur_paren'):
        lspec.cur_paren = []
        lspec.cur_paren.append(ref(lspec.ctype))
    if not hasattr(lspec.cur_paren[-1](), '_params'):
        lspec.cur_paren[-1]()._params = []
    return True

@meta.hook(Declaration)
def add_param(self, lspec, param):
    lspec.cur_paren[-1]()._params.append(param.node)
    return True

@meta.hook(Declaration)
def add_ellipsis(self, lspec):
    lspec.cur_paren[-1]()._ellipsis = True
    return True

@meta.hook(Declaration)
def new_blockinit(self, init_list):
    init_list.node = nodes.BlockInit([])
    return True

@meta.hook(Declaration)
def new_blockexpr(self, ast, current_block):
    ast.node = nodes.BlockExpr([])
    current_block.node = ast.node
    parent = self.rulenodes.parents
    if 'current_block' in parent:
        current_block.node.types = parent['current_block'].node.types.new_child()
    return True

@meta.hook(Declaration)
def add_init(self, ast, expr, designation):
    ast.node.body.append(expr.node)
    if designation.value != "":
        ast.node.body[-1].designation = designation.value
    return True

@meta.hook(Declaration)
def for_decl_begin(self, current_block):
    current_block.node = Node()
    current_block.node.body = []
    # new to link this fake body to other block
    parent = self.rulenodes.parents
    if 'current_block' in parent:
        current_block.node.types = parent['current_block'].node.types.new_child()
    return True

@meta.hook(Declaration)
def for_decl_end(self, init, current_block):
    init.node = None
    if len(current_block.node.body) > 0:
        init.node = current_block.node.body[0]
    return True

@meta.hook(Declaration)
def to_cast(self, ast, typename):
    ast.node = nodes.Cast(nodes.Raw('()'), [typename.node.ctype, ast.node])
    return True

@meta.hook(Declaration)
def sizeof(self, ident):
    if ident.value in Idset and Idset[ident.value] == "sizeof":
        return True
    return False

@meta.hook(Declaration)
def new_sizeof(self, ast, i, n):
    thing = n.node
    if isinstance(thing, nodes.Decl):
        thing = n.node.ctype
    ast.node = nodes.Sizeof(nodes.Raw(i.value), [thing])
    return True

@meta.hook(Declaration)
def new_builtoffset(self, ast, bof):
    ast.node = nodes.Raw("__builtin_offsetof(" + bof.value + ")")
    return True



from pyrser import meta, directives
from pyrser.hooks import copy, echo
from pyrser.parsing.node import Node
from cnorm import nodes
from pyrser.grammar import Grammar
from cnorm.parsing.literal import Literal

# TODO: handle dialect differently
Idset = {
            # ASM
            "asm"                   : "asm",
            "__asm"                 : "asm", 
            "__asm__"               : "asm",
            # ATTRIBUTE
            "attribute"             : "attribute", 
            "__attribute"           : "attribute",
            "__attribute__"         : "attribute",
            #      windows attribute 
            "__declspec"            : "attribute",
            ############
            "auto"                  : "qualifier",
            # SPECIFIER
            #      sign
            "unsigned"              : "sign_unsigned", 
            "__unsigned"            : "sign_unsigned",
            "__unsigned__"          : "sign_unsigned",
            "signed"                : "sign_signed",
            "__signed"              : "sign_signed",
            "__signed__"            : "sign_signed",
            #      size
            "short"                 : "specifier_size",
            "long"                  : "specifier_size_size",
            #      composed
            "struct"                : "specifier_block",
            "union"                 : "specifier_block",
            "enum"                  : "specifier_enum",
            #      function
            "inline"                : "funspecifier",
            "__inline"              : "funspecifier",
            "__inline__"            : "funspecifier",
            "__forceinline"         : "funspecifier",
            # STORAGE
            "register"              : "storage",
            "typedef"               : "storage",
            "static"                : "storage",
            "extern"                : "storage",
            "__thread"              : "storage",
            # QUALIFIER
            "const"                 : "qualifier",
            "volatile"              : "qualifier",
            "restrict"              : "qualifier",
            "__volatile"            : "qualifier",
            "__volatile__"          : "qualifier",
            "__const"               : "qualifier",
            "__restrict"            : "qualifier",
            #      windows qualifier 
            "__cdecl"               : "qualifier",
            "__stdcall"             : "qualifier",
            "__fastcall"            : "qualifier",
            "__w64"                 : "qualifier",
            "__ptr32"               : "qualifier",
            "__ptr64"               : "qualifier",
            # TYPE
            "void"                  : "type",
            "char"                  : "type",
            "int"                   : "type",
            "float"                 : "type",
            "double"                : "type",
            "_Complex"              : "type",
            "__complex"             : "type",
            "__complex__"           : "type",
            "_Imaginary"            : "type",
            "_Bool"                 : "type",
            "__label__"             : "type",
            "__builtin_va_list"     : "type",
            "__int8"                : "type",
            "__int16"               : "type",
            "__int32"               : "type",
            "__int64"               : "type",
            # SIZEOF/TYPEOF
            #      sizeof
            "sizeof"                : "sizeof",
            "__alignof"             : "sizeof",
            "__alignof__"           : "sizeof",
            #      typeof
            "typeof"                : "typeof",
            "__typeof"              : "typeof",
            "__typeof__"            : "typeof",
            # UNARY OP
            "__real"                : "unary",
            "__real__"              : "unary",
            "__imag"                : "unary",
            "__imag__"              : "unary",
            "__extension__"         : "unary",
            # RESERVED KEYWORDS
            "switch"                : "reserved",
            "case"                  : "reserved",
            "default"               : "reserved",
            "if"                    : "reserved",
            "else"                  : "reserved",
            "while"                 : "reserved",
            "do"                    : "reserved",
            "for"                   : "reserved",
            "goto"                  : "reserved",
            "continue"              : "reserved",
            "break"                 : "reserved",
            "return"                : "reserved",
            "__builtin_offsetof"    : "reserved"
}

class Expression(Grammar, Literal):
    """
        interaction with other CNORM PART:

    """
    entry = "expression"
    grammar = """

        /*
            Comment works as in C/C++
        */

        dummy_with_brace ::= @ignore('null')
            [
            '{' dummy_with_brace* '}'
            | Base.string
            | Base.char
            | Base.read_char:c #check_not_brace(c)
            ]
        ;

        dummy_with_paren ::= @ignore('null')
            [
            '(' dummy_with_paren* ')'
            | Base.string
            | Base.char
            | Base.read_char:c #check_not_paren(c)
            ]
        ;

        expression ::=
            assignement_expression:_
            [
                ',':op #new_raw(op, op)
                assignement_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        assign_op ::= @ignore('null') ['=' !'='| "+=" | "-=" | "*=" | "/=" | "%=" | "<<=" | ">>=" | "&=" | "^=" | "|="]:op #new_raw(_, op);
        assignement_expression ::=
            conditional_expression:_
            [
                assign_op:op
                assignement_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        constant_expression ::=
            conditional_expression:_
        ;

        conditional_expression ::=
            logical_or_expression:_
            [
                '?'
                expression?:then
                ':'
                assignement_expression?:else
                #new_ternary(_, then, else)
            ]?
        ;

        logical_or_op ::= @ignore('null') ["||"]:op #new_raw(_, op);
        logical_or_expression ::=
            logical_and_expression:_
            [   
                logical_or_op:op
                logical_and_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        logical_and_op ::= @ignore('null') ["&&"]:op #new_raw(_, op);
        logical_and_expression ::=
            or_expression:_
            [   
                logical_and_op:op
                or_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        or_op ::= @ignore('null') ["|" !["|"|"="]]:op #new_raw(_, op);
        or_expression ::=
            xor_expression:_
            [   
                or_op:op
                xor_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        xor_op ::= @ignore('null') ["^" !"="]:op #new_raw(_, op);
        xor_expression ::=
            and_expression:_
            [   
                xor_op:op
                and_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        and_op ::= @ignore('null') ["&" !["&"|"="]]:op #new_raw(_, op);
        and_expression ::=
            equality_expression:_
            [   
                and_op:op
                equality_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        eqneq_op ::= @ignore('null') ["==" | "!="]:op #new_raw(_, op);
        equality_expression ::=
            relational_expression:_
            [   
                eqneq_op:op
                relational_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        cmp_op ::= @ignore('null') ["<=" | ">=" | '<' !'<' | '>' !'>' ]:op #new_raw(_, op);
        relational_expression ::=
            shift_expression:_
            [   
                cmp_op:op
                shift_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        shift_op ::= @ignore('null') ["<<" !"=" | ">>" !"="]:op #new_raw(_, op);
        shift_expression ::=
            additive_expression:_
            [   
                shift_op:op
                additive_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        add_op ::= @ignore('null') ['+' !['+'|'='] | '-' !['-'|'='|'>']]:op #new_raw(_, op);
        additive_expression ::=
            multiplicative_expression:_
            [   
                add_op:op
                multiplicative_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        mul_op ::= @ignore('null') [['*'|'/'|'%']:op !'='] #new_raw(_, op);
        multiplicative_expression ::=
            unary_expression:_
            [   
                mul_op:op
                unary_expression:param
                #new_binary(_, op, param)
            ]*
        ;

        unary_op ::= @ignore('null') 
            ["++"
            |"--"
            |"&&"
            | '&' !'='
            | '*' !'='
            | '~' !'='
            | '!' !'='
            | '+' !'='
            | '-' !['>'|'=']
            ]:op #new_raw(_, op)
        ;
        unary_expression ::=
            postfix_expression:_
            |
            [   unary_op:op
            |   Base.id:i
                #is_unary(op, i)
            ]:op
            unary_expression:expr
            #new_unary(_, op, expr)
        ;

        postfix_expression ::=
            primary_expression:_
            [
                '[' expression:expr ']' #new_array_call(_, _, expr)
                | '(' func_arg_list?:args ')' #new_func_call(_, _, args)
                | '.' identifier:i #new_dot(_, _, i)
                | "->" identifier:i #new_arrow(_, _, i)
                | ["++"|"--"]:op #new_raw(op, op) #new_post(_, op, _)
            ]*
        ;

        func_arg_list ::=
            assignement_expression:a #new_arg(_, a)
            [   ','
                assignement_expression:a #new_arg(_, a)
            ]*
        ;

        primary_expression ::=
            '(' expression:expr ')' #new_paren(_, expr)
            | [Literal.literal
            | identifier]:_
        ;

        identifier ::=
            @ignore('null') 
            [
                rootidentifier:id
                #check_is_id(id)
                #new_id(_, id)
            ]
        ;

        rootidentifier ::= Base.id:_
        ;

    """

    def parse(self, source, entry=None):
        res = Grammar.parse(self, source, entry)
        if hasattr(res, 'node'):
            return res.node
        return res

@meta.hook(Expression)
def new_ternary(self, ast, then_expr, else_expr):
    then_body = None
    if hasattr(then_expr, 'node'):
        then_body = then_expr.node
    else_body = None
    if hasattr(else_expr, 'node'):
        else_body = else_expr.node
    ast.node = nodes.Ternary([], [ast.node, then_body, else_body])
    return True

@meta.hook(Expression)
def new_binary(self, ast, op, param):
    ast.node = nodes.Binary(op.node, [ast.node, param.node])
    return True

@meta.hook(Expression)
def is_unary(self, op, ident):
    if ident.value in Idset and Idset[ident.value] == "unary":
        op.node = nodes.Raw(ident.value + " ")
        return True
    return False

@meta.hook(Expression)
def new_unary(self, ast, op, param):
    ast.node = nodes.Unary(op.node, [param.node])
    return True

@meta.hook(Expression)
def new_paren(self, ast, expr):
    ast.node = nodes.Paren("()", [expr.node])
    return True

@meta.hook(Expression)
def new_post(self, ast, op, param):
    ast.node = nodes.Post(op.node, [param.node])
    return True

@meta.hook(Expression)
def new_arg(self, ast, arg):
    if not hasattr(ast, 'node'):
        ast.node = []
    ast.node.append(arg.node)
    return True

@meta.hook(Expression)
def new_array_call(self, ast, call, index):
    ast.node = nodes.Array(call.node, [index.node])
    return True

@meta.hook(Expression)
def new_dot(self, ast, call, field):
    ast.node = nodes.Dot(call.node, [field.node])
    return True

@meta.hook(Expression)
def new_arrow(self, ast, call, field):
    ast.node = nodes.Arrow(call.node, [field.node])
    return True

@meta.hook(Expression)
def new_func_call(self, ast, call, args):
    if hasattr(args, 'node'):
        ast.node = nodes.Func(call.node, args.node)
    else:
        ast.node = nodes.Func(call.node, [])
    return True

@meta.hook(Expression)
def new_raw(self, ast, data):
    ast.node = nodes.Raw(data.value)
    return True

@meta.hook(Expression)
def new_id(self, ast, identifier):
    ast.node = nodes.Id(identifier.value)
    return True

@meta.hook(Expression)
def check_not_brace(self, c):
    return c.value != "{" and c.value != "}" and c.value != "'" and c.value != '"'

@meta.hook(Expression)
def check_not_paren(self, c):
    return c.value != "(" and c.value != ")" and c.value != "'" and c.value != '"'

@meta.hook(Expression)
def check_is_id(self, identifier):
    return identifier.value not in Idset



#!/usr/bin/env python3
"""
        Type handling
"""

# scope contient toutes les surcharges
# f1(int,int)int ou f1(double, char)truc
# a as int ou a as double
# b as int ou b as toto
# c as int ou c as toto ou c as tata
scope = { 'f1': [
                {'symbol':'f1_int_int_int', 'param':['int', 'int'], 'type':'int'},
                {'symbol':'f1_truc_double_char', 'param':['double', 'char'], 'type':'truc'},
                ],
          'a':[{'type':'int', 'symbol':'a_int'}, {'type':'double', 'symbol':'a_double'}],
          'b':[{'type':'int', 'symbol':'b_int'}, {'type':'toto', 'symbol':'b_toto'}], 
          'c':[{'type':'int', 'symbol':'c_int'}, {'type':'tata', 'symbol':'c_tata'}, {'type':'toto', 'symbol':'c_toto'}]
        }
print("scope: %s" % repr(scope))

class   Log:
        indent = 0
        def     incdent():
                Log.indent += 1
        def     decdent():
                Log.indent -= 1
        def     log(s):
                print("%s%s" % ("\t" * Log.indent, s))

class   THandling:
        def     __init__(self, scope):
                self.scope = scope
        
        # retro propagation des types
        def     typeFeedback(self, expr, expr_type):
                Log.log("type this expr with symbol %s" % expr_type)
                expr['symbol'] = expr_type['symbol']
                expr['type'] = expr_type['type']
                # before remove candidates retro propagate parameter
                if 'param' in expr_type:
                        fun_arity = len(expr_type['param'])
                        for pidx in range(0, fun_arity):
                                Log.log("at INDEX %d = %s" % (pidx, repr(expr['param'])))
                                param_candidat = {tp['type']:tp for tp in expr['param'][pidx]['candidates']}
                                if expr_type['param'][pidx] in param_candidat:
                                        Log.incdent()
                                        self.typeFeedback(expr['param'][pidx], param_candidat[expr_type['param'][pidx]])
                                        Log.decdent()
                del expr['candidates']

        # type une expression
        def     typeExpr(self, expr):
                # result
                candidates = []
                Log.log("expr: %s" % repr(expr))
                # search all overloads for the function
                overloads = self.scope[expr['name']]
                Log.log("found overloads for function %s" % expr['name'])
                # get all thing overloads
                for thing in overloads:
                        Log.log("overload %s" % thing['symbol'])
                        # is just a var
                        if 'param' not in expr or len(expr['param']) == 0:
                                Log.log("Is JUST A THING:")
                                candidates.append(thing)
                        else:
                                # count function arity in expression
                                fun_arity = len(expr['param'])
                                # select only function overloads with same arity TODO: ellipsis
                                if (len(thing['param']) == fun_arity):
                                        Log.log("select %s for arity" % thing['symbol'])
                                        # result list param
                                        list_param = []
                                        # for all param fetch possible type
                                        for pidx in range(0, fun_arity):
                                                Log.log("idx: %s" % pidx)
                                                # get name of param
                                                pname = expr['param'][pidx]['name']
                                                # get types of this param
                                                Log.incdent()
                                                self.typeExpr(expr['param'][pidx])
                                                Log.decdent()
                                                types = set([tp['type'] for tp in expr['param'][pidx]['candidates']])
                                                Log.log("param %s is %s" % (pname, repr(types)))
                                                # intersect type of overload and types
                                                ptype = thing['param'][pidx]
                                                Log.log("intersect with %s" % ptype)
                                                if ptype in types:                                        
                                                        Log.log("FOUND: %s" % ptype)
                                                        list_param.append(ptype)
                                        # test the candidates
                                        if (len(list_param) == fun_arity):
                                                candidates.append(thing)
                # add candidates in expression
                if len(candidates) > 0:
                        expr['candidates'] = candidates
                        # TODO: retro propagation
                        if len(candidates) == 1:
                                Log.log("retro propagation")
                                Log.incdent()
                                self.typeFeedback(expr, expr['candidates'][0])
                                Log.decdent()
                        return True
                return False

typehandler = THandling(scope)

# expression
expr = {'name':'f1', 'param':[{'name':'a'}, {'name':'b'}]}
print("EXPR BEFORE:%s" % repr(expr))
# call type handling
if typehandler.typeExpr(expr):
        print("EXPR AFTER: %s" % repr(expr))

if __name__ == "__main__":
        import doctest
        doctest.testmod(verbose=True)



from pyrser import parsing, meta

Storages = meta.enum('AUTO', 'REGISTER', 'TYPEDEF', 'STATIC', 'EXTERN', 'INLINE', 'VIRTUAL', 'EXPLICIT', 'FORCEINLINE', 'THREAD')
Qualifiers = meta.enum('AUTO', 'CONST', 'VOLATILE', 'RESTRICT', 'W64', 'STDCALL', 'CDECL', 'PTR32', 'PTR64', 'FASTCALL')
Specifiers = meta.enum('AUTO', 'STRUCT', 'UNION', 'ENUM', 'LONG',
                       'LONGLONG', 'SHORT')
Signs = meta.enum('AUTO', 'SIGNED', 'UNSIGNED')

# EXPRESSION PART

class Expr(parsing.Node):
    """All expression"""

class Func(Expr):
    """For almost everything"""

    def __init__(self, call_expr: Expr, params: list):
        Expr.__init__(self)
        self.call_expr = call_expr
        self.params = params

class BlockInit(Expr):
    """Initializer Block Expression"""

    def __init__(self, body: [Expr]):
        self.body = body

class BlockExpr(Expr):
    """Compound Block Expression"""

    def __init__(self, body: [Expr]):
        self.body = body

class Unary(Func):
    """For unary operator"""

class Paren(Unary):
    """For () expression"""

class Array(Unary):
    """For [] expression"""

class Dot(Unary):
    """For . expression"""

class Arrow(Unary):
    """For -> expression"""

class Post(Unary):
    """For post{inc,dec} expression"""

class Sizeof(Unary):
    """For sizeof expr/type expression"""

class Binary(Func):
    """For binary operator"""

class Cast(Binary):
    """For cast operator"""

class Range(Binary):
    """For range expression"""

class Ternary(Func):
    """For ternary operator"""

class Terminal(Expr):
    """For Terminal expression"""

    def __init__(self, value: str):
        Expr.__init__(self)
        self.value = value

class Id(Terminal):
    """Terminal Id"""

class Literal(Terminal):
    """Terminal Literal"""

class Raw(Terminal):
    """Terminal Raw"""


# DECLARATION PART

class Enumerator(parsing.Node):
    """Enumerator A=x in enums"""

    def __init__(self, ident: str, expr: Expr):
        self.ident = ident
        self.expr = expr

class DeclType(parsing.Node):
    """For type in declaration"""
    
    def __init__(self):
        self._decltype = None

    def link(self, t: 'DeclType'=None):
        if t != None:
            if not isinstance(t, DeclType):
                raise Exception("add only C type declarator")
            self._decltype = t
        return self._decltype

    def push(self, t: 'DeclType'=None):
        if t != None:
            if not isinstance(t, DeclType):
                raise Exception("add only C type declarator")
            old = self._decltype
            self._decltype = t
            self._decltype.link(old)
        return self._decltype


class PointerType(DeclType):
    """For pointer in declaration"""


class ArrayType(DeclType):
    """For array in declaration"""

    def __init__(self, expr=None):
        DeclType.__init__(self)
        self._expr = expr
    
    @property
    def expr(self):
        return self._expr

class ParenType(DeclType):
    """For parenthesis in declaration"""

    def __init__(self, params=None):
        DeclType.__init__(self)
        if params == None:
            params = []
        self._params = params

    @property
    def params(self):
        return self._params

class QualType(DeclType):
    """For qualifier in declaration"""

    def __init__(self, qualifier: Qualifiers= Qualifiers.AUTO):
        DeclType.__init__(self)
        # qualifier in (auto, const, volatile, restrict)
        self._qualifier = qualifier

class AttrType(DeclType):
    """For attribute specifier in declaration"""

    def __init__(self, raw: str):
        DeclType.__init__(self)
        self._attr = raw


class CType(parsing.Node):
    """Base for primary/func
        
        TODO: 
            - add a function that iter decltypes and delete unnecessary parenthesis
    """

    def __init__(self):
        parsing.Node.__init__(self)
        self._decltype = None
        # only one storage by declaration (auto, register, typedef, static, extern, ...)
        self._storage = Storages.AUTO
        # only one specifier by declaration (auto, short, long, struct, union, enum)
        self._specifier = Specifiers.AUTO

    def copy(self):
        import copy
        theclone = copy.copy(self)
        theclone._decltype = None
        return theclone


    def link(self, t: DeclType=None):
        if t != None:
            if not isinstance(t, DeclType):
                raise Exception("add only C type declarator")
            self._decltype = t
        return self._decltype

    def push(self, t: DeclType=None):
        if t != None:
            if not isinstance(t, DeclType):
                raise Exception("add only C type declarator")
            old = self._decltype
            self._decltype = t
            self._decltype.link(old)
        return self._decltype

class PrimaryType(CType):
    """For primary type in declaration"""

    def __init__(self, identifier: str):
        CType.__init__(self)
        # identifier (void, char, int, float, double, typedefname)
        self._identifier = identifier

    @property
    def identifier(self):
        return self._identifier

class ComposedType(CType):
    """For composed type in declaration"""

    def __init__(self, identifier: str):
        CType.__init__(self)
        # identifier (name of the struct/union/enum)
        self._identifier = identifier
        # if struct
        #self.fields = []
        # if enum
        #self.enums = []

    @property
    def identifier(self):
        return self._identifier


class FuncType(PrimaryType):
    """For function in declaration"""

    def __init__(self, identifier: str, params=[], decltype=None):
        PrimaryType.__init__(self, identifier)
        self.opened = True
        if decltype != None:
            self._decltype = decltype
        self._params = params
        # when defined
        #self.body = BlockStmt([])

    @property
    def params(self):
        return self._params

# helper to create a CType from previous one
def makeCType(declspecifier: str, ctype=None):
    from cnorm.parsing.expression import Idset
    #print("%s %s" %(declspecifier, ctype))
    if ctype == None:
        ctype = PrimaryType('int')
    if Idset[declspecifier] == "type":
        ctype._identifier = declspecifier
    if Idset[declspecifier] == "storage":
        cleantxt = declspecifier.strip("_")
        ctype._storage = Storages.map[cleantxt.upper()]
    if Idset[declspecifier] == "qualifier":
        cleantxt = declspecifier.strip("_")
        ctype.link(QualType(Qualifiers.map[cleantxt.upper()]))
    if Idset[declspecifier] == "funspecifier":
        cleantxt = declspecifier.strip("_")
        ctype._storage = Storages.map[cleantxt.upper()]
    if Idset[declspecifier] == "sign_unsigned":
        cleantxt = declspecifier.strip("_")
        ctype._sign = Signs.map[cleantxt.upper()]
    if Idset[declspecifier] == "sign_signed":
        cleantxt = declspecifier.strip("_")
        ctype._sign = Signs.map[cleantxt.upper()]
    if Idset[declspecifier] == "specifier_block":
        cleantxt = declspecifier.strip("_")
        ctype._specifier = Specifiers.map[cleantxt.upper()]
    if Idset[declspecifier] == "specifier_enum":
        cleantxt = declspecifier.strip("_")
        ctype._specifier = Specifiers.map[cleantxt.upper()]
    if Idset[declspecifier] == "specifier_size":
        cleantxt = declspecifier.strip("_")
        ctype._specifier = Specifiers.map[cleantxt.upper()]
    if Idset[declspecifier] == "specifier_size_size":
        cleantxt = declspecifier.strip("_")
        if ctype._specifier == Specifiers.AUTO:
            ctype._specifier = Specifiers.map[cleantxt.upper()]
        else:
            ctype._specifier = Specifiers.LONGLONG
    return ctype

class Decl(Expr):
    """For basic declaration"""

    def __init__(self, name: str, ct=None):
        if ct == None:
            ct = PrimaryType('int')
        Expr.__init__(self)
        self._name = name
        self._ctype = ct

    @property
    def ctype(self) -> CType:
        return self._ctype

    def assign_expr(self, expr=None):
        if not hasattr(self, '_assign_expr'):
            self._assign_expr = None
        if expr != None:
            self._assign_expr = expr
        return self._assign_expr

    def colon_expr(self, expr=None):
        if not hasattr(self, '_colon_expr'):
            self._colon_expr = None
        if expr != None:
            self._colon_expr = expr
        return self._colon_expr

# STATEMENT PART

class Stmt(parsing.Node):
    """For statement"""

class ExprStmt(Stmt):
    """Expression statement"""

    def __init__(self, expr: Expr):
        parsing.Node.__init__(self)
        self.expr = expr

class BlockStmt(Stmt):
    """Block statement"""

    def __init__(self, body: [ExprStmt]):
        parsing.Node.__init__(self)
        self.body = body

    def func(self, name: str):
        """return the func defined named name"""

    def var(self, name: str):
        """return the var instancied named name"""

    def type(self, name: str):
        """return the complete definition of type 'name'"""

    def declfuncs(self, name: str):
        """return all declaration of function 'name'"""

    def declvars(self, name: str):
        """return all declaration of variable 'name'"""

    def decltypes(self, name: str):
        """return all declaration of type 'name'"""

class RootBlockStmt(BlockStmt):
    """Root Block statement"""

    def __init__(self, body: [ExprStmt]):
        BlockStmt.__init__(self, body)
        from collections import ChainMap
        self.types = ChainMap()


class Label(Stmt):
    """Label statement"""

    def __init__(self, value: str):
        Stmt.__init__(self)
        self.value = value

class Branch(Label):
    """branch statement"""

    def __init__(self, value:str, expr: Expr):
        Label.__init__(self, value)
        self.expr = expr

class Case(Branch):
    """Case statement"""

    def __init__(self, expr: Expr):
        Branch.__init__(self, "case", expr)

class Return(Branch):
    """Return statement"""

    def __init__(self, expr: Expr):
        Branch.__init__(self, "return", expr)

class Goto(Branch):
    """Goto statement"""

    def __init__(self, expr: Expr):
        Branch.__init__(self, "goto", expr)

class LoopControl(Label):
    """loop control statement"""

class Break(LoopControl):
    """break statement"""

    def __init__(self):
        Label.__init__(self, "break")

class Continue(LoopControl):
    """continue statement"""

    def __init__(self):
        Label.__init__(self, "continue")

class Conditional(Stmt):
    """Conditional statement"""

    def __init__(self, condition: Expr):
        Stmt.__init__(self)
        self.condition = condition

class If(Conditional):
    """If statement"""

    def __init__(self, condition: Expr, thencond: Stmt, elsecond: Stmt=None):
        Conditional.__init__(self, condition)
        self.thencond = thencond
        self.elsecond = elsecond

class While(Conditional):
    """While statement"""
    
    def __init__(self, condition: Expr, body: Stmt):
        Conditional.__init__(self, condition)
        self.body = body

class Switch(Conditional):
    """Switch statement"""
    
    def __init__(self, condition: Expr, body: Stmt):
        Conditional.__init__(self, condition)
        self.body = body

class Do(Conditional):
    """Do statement"""

    def __init__(self, condition: Expr, body: Stmt):
        Conditional.__init__(self, condition)
        self.body = body

class For(Stmt):
    """For statement"""

    def __init__(self, init: Expr, condition: Expr, increment: Expr, body: Stmt):
        Stmt.__init__(self)
        self.init= init
        self.condition = condition
        self.increment = increment
        self.body = body




from pyrser import meta
from pyrser import fmt
from pyrser.passes import to_yml
from cnorm import nodes


# DECL

@meta.add_method(nodes.CType)
def ctype_to_c(self, func_var_name=""):
    # our global declarator
    declarator = fmt.sep("", [])
    # typename or full decl
    if func_var_name != "":
        declarator.lsdata.append(func_var_name)
    # intern prototype
    if hasattr(self, 'params'):
        # param list
        pf = fmt.sep(", ", [])
        for p in self.params:
            if p.ctype != None:
                if isinstance(p.ctype, nodes.CType):
                    pf.lsdata.append(p.ctype.ctype_to_c(p._name))
        if hasattr(self, '_ellipsis'):
            pf.lsdata.append('...')
        if len(pf.lsdata) > 0:
            declarator.lsdata.append(fmt.block('(', ')',  pf))
        else:
            declarator.lsdata.append('()')
    # for externalize the last qualifier
    qualextern = None
    # final output
    decl_ls = fmt.sep(" ", [])
    if self.link() != None:
        # add all qualifiers
        if len(declarator.lsdata) > 0:
            qual_list = declarator
        else:
            qual_list = fmt.sep(" ", [])
        unqual_list = self.link()
        # qualification of declaration
        while unqual_list != None:
            if isinstance(unqual_list, nodes.ParenType):
                # surround previous defs by ()
                qual_list = fmt.sep("", [fmt.block("(", ")", [qual_list])])
                # () provide param for function pointers
                if len(unqual_list.params) > 0:
                    pf = fmt.sep(", ", [])
                    for p in unqual_list.params:
                         pf.lsdata.append(p.ctype.ctype_to_c(p._name))
                    if hasattr(unqual_list, '_ellipsis'):
                        pf.lsdata.append('...')
                    qual_list.lsdata.append(fmt.block('(', ')', pf))
            if isinstance(unqual_list, nodes.PointerType):
                qual_list.lsdata.insert(0, "*")
            if isinstance(unqual_list, nodes.AttrType):
                qual_list.lsdata.insert(0, unqual_list._attr + " ")
            if isinstance(unqual_list, nodes.QualType):
                if unqual_list._qualifier != nodes.Qualifiers.AUTO:
                    if unqual_list.link() == None:
                        qualextern = unqual_list
                    else:
                        qual_list.lsdata.insert(0, nodes.Qualifiers.rmap[unqual_list._qualifier].lower() + " ")
            if isinstance(unqual_list, nodes.ArrayType):
                # collect all consecutive array
                consec_ary = []
                consec_ary.append(unqual_list)
                unqual_list = unqual_list.link()
                while unqual_list != None and isinstance(unqual_list, nodes.ArrayType):
                    consec_ary.append(unqual_list)
                    unqual_list = unqual_list.link()
                reordered = []
                for ary in consec_ary:
                    if ary.expr != None:
                        reordered.insert(0, fmt.block("[", "]", [ary.expr.to_c()]))
                    else:
                        reordered.insert(0, "[]")
                qual_list.lsdata.extend(reordered)
                # rewind one for last sentence
                unqual_list = consec_ary[-1]
            unqual_list = unqual_list.link()
        # add qualified declarator
        decl_ls.lsdata.append(qual_list)
    elif len(declarator.lsdata) > 0:
        # no qualifiers just the name
        decl_ls.lsdata.append(declarator)
    # for enum
    if hasattr(self, 'enums'):
        enums = fmt.sep(",\n", [])
        for enum in self.enums:
            if enum.expr != None:
                enums.lsdata.append(fmt.sep(" = ", [enum.ident, enum.expr.to_c()]))
            else:
                enums.lsdata.append(enum.ident)
        decl_ls.lsdata.insert(0, fmt.tab(fmt.block("{\n", "}", enums)))
    # for struct
    if hasattr(self, 'fields'):
        fields = []
        for field in self.fields:
            fields.append(field.to_c())
        decl_ls.lsdata.insert(0, fmt.tab(fmt.block("{\n", "}", fields)))
    # just the type name
    if hasattr(self, 'identifier'):
        decl_ls.lsdata.insert(0, self.identifier)
    # attributes composed
    if hasattr(self, '_attr_composed'):
        decl_ls.lsdata.insert(0, fmt.sep(" ", self._attr_composed))
    # specifier
    if self._specifier != nodes.Specifiers.AUTO:
        if self._specifier == nodes.Specifiers.LONGLONG:
            decl_ls.lsdata.insert(0, "long long")
        else:
            decl_ls.lsdata.insert(0, nodes.Specifiers.rmap[self._specifier].lower())
    # sign
    if hasattr(self, '_sign') and self._sign != nodes.Signs.AUTO:
        decl_ls.lsdata.insert(0, nodes.Signs.rmap[self._sign].lower())
    # qualifier externalized
    if qualextern != None:
        decl_ls.lsdata.insert(0, nodes.Qualifiers.rmap[qualextern._qualifier].lower())
    # End by storage
    if self._storage != nodes.Storages.AUTO:
        decl_ls.lsdata.insert(0, nodes.Storages.rmap[self._storage].lower())
    return decl_ls

def decl_to_c(self):
    decl_ls = self.ctype.ctype_to_c(self._name)
    # add bitfield
    if hasattr(self, '_colon_expr'):
        decl_ls.lsdata.append(":")
        decl_ls.lsdata.append(self._colon_expr.to_c())
    # attributes decl
    if hasattr(self, '_attr_decl'):
        decl_ls.lsdata.extend(self._attr_decl)
    # add initializers
    if hasattr(self, '_assign_expr'):
        decl_ls.lsdata.append("=")
        decl_ls.lsdata.append(self._assign_expr.to_c())
    return decl_ls

@meta.add_method(nodes.Decl)
def to_c(self):
    if hasattr(self, 'body') and self.body != None:
        return fmt.sep("\n", [self.ctype.ctype_to_c(self._name), self.body.to_c()])
    else:
        return fmt.end(';\n', decl_to_c(self))

# STATEMENT

@meta.add_method(nodes.For)
def to_c(self):
    init_body = None
    if type(self.init) is nodes.Decl:
        init_body = decl_to_c(self.init)
    elif self.init != None:
        init_body = self.init.expr.to_c()
    cond_body = None
    if self.condition != None:
        cond_body = self.condition.expr.to_c()
    inc_body = None
    if self.increment != None:
        inc_body = self.increment.to_c()
    lsfor = [
                fmt.sep(" ", ["for", fmt.block('(', ')\n', 
                    [fmt.sep("; ",
                        [
                            init_body,
                            cond_body,
                            inc_body
                        ])
                    ])]),
                fmt.tab(self.body.to_c())
            ]
    return fmt.end("", lsfor)

@meta.add_method(nodes.If)
def to_c(self):
    thenbody = ';\n'
    if self.thencond != None:
        thenbody = fmt.tab(self.thencond.to_c())
    lsif = [
                fmt.sep(" ", ["if", fmt.block('(', ')\n', [self.condition.to_c()])]),
                thenbody
            ]
    if self.elsecond != None:
        lsif.append("else\n")
        lsif.append(fmt.tab(self.elsecond.to_c()))
    return fmt.end("", lsif)

@meta.add_method(nodes.While)
def to_c(self):
    body = ';\n'
    if self.body != None:
        body = fmt.tab(self.body.to_c())
    lswh = [
                fmt.sep(" ", ["while", fmt.block('(', ')\n', [self.condition.to_c()])]),
                body
            ]
    return fmt.sep("", lswh)

@meta.add_method(nodes.Do)
def to_c(self):
    body = ';\n'
    if self.body != None:
        body = fmt.tab(self.body.to_c())
    lsdo = [
                fmt.sep("\n", ["do", body]),
                fmt.sep(" ", ["while", fmt.block('(', ');\n', [self.condition.to_c()])]),
           ]
    return fmt.sep("", lsdo)

@meta.add_method(nodes.Switch)
def to_c(self):
    body = ';\n'
    if self.body != None:
        body = fmt.tab(self.body.to_c())
    lswh = [
                fmt.sep(" ", ["switch", fmt.block('(', ')\n', [self.condition.to_c()])]),
                body
            ]
    return fmt.sep("", lswh)

@meta.add_method(nodes.Label)
def to_c(self):
    return fmt.end(":\n", [self.value])

@meta.add_method(nodes.LoopControl)
def to_c(self):
    return fmt.end(";\n", [self.value])

@meta.add_method(nodes.Branch)
def to_c(self):
    body = ';\n'
    if self.expr != None:
        body = self.expr.to_c()
    return fmt.end(";\n", [fmt.sep(" ", [self.value, body])])

@meta.add_method(nodes.Case)
def to_c(self):
    return fmt.end(":\n", [fmt.sep(" ", [self.value, self.expr.to_c()])])

@meta.add_method(nodes.ExprStmt)
def to_c(self):
    return fmt.end(";\n", [self.expr.to_c()])

@meta.add_method(nodes.BlockStmt)
def to_c(self):
    lsbody = []
    for e in self.body:
        if e != None:
            lsbody.append(e.to_c())
    return fmt.block("{\n", "}\n", fmt.tab(lsbody))

@meta.add_method(nodes.BlockExpr)
def to_c(self):
    lsbody = []
    for e in self.body:
        if e != None:
            lsbody.append(e.to_c())
    return fmt.block("({\n", "})", fmt.tab(lsbody))

@meta.add_method(nodes.RootBlockStmt)
def to_c(self):
    lsbody = []
    for e in self.body:
        lsbody.append(e.to_c())
    return fmt.sep("", lsbody)

# EXPRESSION

@meta.add_method(nodes.Func)
def to_c(self):
    lsparams = []
    # TODO: add forward declarations
    for p in self.params:
        lsparams.append(p.to_c())
    return fmt.sep("", [self.call_expr.to_c(), fmt.block('(', ')', [fmt.sep(', ', lsparams)])])

@meta.add_method(nodes.BlockInit)
def to_c(self):
    lsbody = []
    i = 0
    for it in self.body:
        subbody = it.to_c()
        if hasattr(it, 'designation'):
            subbody = fmt.sep("", [it.designation, subbody])
        if isinstance(subbody, fmt.block):
            lsbody.append(fmt.tab(["\n", subbody]))
        else:
            lsbody.append(subbody)
        i += 1
        if i > 8:
            lsbody[-1] = "\n" + str(lsbody[-1])
            i = 0
    return fmt.block("{ ", " }", fmt.tab(fmt.sep(', ', lsbody)))

@meta.add_method(nodes.Ternary)
def to_c(self):
    condexpr = None
    if self.params[0] != None:
        condexpr = self.params[0].to_c()
    thenexpr = None
    if self.params[1] != None:
        thenexpr = self.params[1].to_c()
    content = fmt.sep("", [condexpr, ' ? ', thenexpr])
    if len(self.params) > 2:
        elseexpr = None
        if self.params[2] != None:
            elseexpr = self.params[2].to_c()
        content.lsdata.append(' : ')
        content.lsdata.append(elseexpr)
    return content

@meta.add_method(nodes.Binary)
def to_c(self):
    lsparams = []
    for p in self.params:
        lsparams.append(p.to_c())
    if type(self.call_expr) is nodes.Raw and self.call_expr.value == ",":
        return fmt.sep(str(self.call_expr.to_c()) + ' ', lsparams)
    return fmt.sep(' ' + str(self.call_expr.to_c()) + ' ', lsparams)

@meta.add_method(nodes.Cast)
def to_c(self):
    return fmt.sep(" ", [fmt.block("(", ")", self.params[0].ctype_to_c()), self.params[1].to_c()])

@meta.add_method(nodes.Unary)
def to_c(self):
    return fmt.sep("", [self.call_expr.to_c(), self.params[0].to_c()])

@meta.add_method(nodes.Sizeof)
def to_c(self):
    if isinstance(self.params[0], nodes.CType):
        return fmt.sep(" ", [self.call_expr.to_c(), fmt.block("(", ")", self.params[0].ctype_to_c())])
    else:
        return fmt.sep(" ", [self.call_expr.to_c(), self.params[0].to_c()])

@meta.add_method(nodes.Dot)
def to_c(self):
    return fmt.sep(".", [self.call_expr.to_c(), self.params[0].to_c()])

@meta.add_method(nodes.Arrow)
def to_c(self):
    return fmt.sep("->", [self.call_expr.to_c(), self.params[0].to_c()])

@meta.add_method(nodes.Paren)
def to_c(self):
    return fmt.block('(', ')', [self.params[0].to_c()])

@meta.add_method(nodes.Array)
def to_c(self):
    return fmt.sep("", [self.call_expr.to_c(), fmt.block('[', ']', [self.params[0].to_c()])])

@meta.add_method(nodes.Post)
def to_c(self):
    return fmt.sep("", [self.params[0].to_c(), self.call_expr.to_c()])

@meta.add_method(nodes.Terminal)
def to_c(self):
    return self.value
